@@ -2639,6 +2639,8 @@ ResolveOneParam(QueryBuild *qb)
 			break;
 
 		case SQL_C_WCHAR:
+            if (SQL_NTS == used)
+                used = WCLEN * wcslen((SQLWCHAR *) buffer);
 			buf = allocbuf = ucs2_to_utf8((SQLWCHAR *) buffer, used / WCLEN, (UInt4 *) &used, FALSE);
 			used *= WCLEN;
 			break;
@@ -2989,6 +2991,11 @@ ResolveOneParam(QueryBuild *qb)
 			if (param_sqltype == SQL_BIT)
 				CVT_APPEND_CHAR(qb, '\'');		/* Open Quote */
 
+			if (param_sqltype == SQL_SMALLINT)
+			{
+				CVT_APPEND_STR(qb, "(");
+			}
+
 			if (buf)
 			{
 				switch (used)
@@ -3008,7 +3015,7 @@ ResolveOneParam(QueryBuild *qb)
 			if (param_sqltype == SQL_SMALLINT)
 			{
 				/* needs cast because there is no automatic downcast from int4 constants */
-				CVT_APPEND_STR(qb, "::int2");
+				CVT_APPEND_STR(qb, ")::int2");
 			}
 
 			if (param_sqltype == SQL_BIT)
